; -------------------------------------
;   Z80 Return-oriented Virtual Machine
;   (c) 2016 - 2020 kio@little-bat.de
; -------------------------------------


.local


; ==============================
;     Heap Memory Management
; ==============================


.if 0

  Memory Map:

    $FFFF +---------------+
    	  |               |
    	  |      Rom      |
    	  |               |
    ~C000 +---------------+
    	  | Resident Data |
          +---------------+ <- handles_end
    	  |               |
    	  |    Handles    | <- free_handle
    	  |______         |
    	         \________| <- handles_start

    	   Free Memory
    	   ______
    	  |      \________  <- heap_end
    	  |               |
    	  | Malloced Data |
    	  |               |
          +---------------+ <- heap_start
    	  | Resident Data |
    ~3000 +---------------+
    	  | Stack? _______|
    	  |_______/       |
    	  |               |
    	  | Application   |
    	  |               |
    ~1000 +---------------+
    	  | Globals       |
    	  | Irpt Handler  |
    	  | RST Vectors   |	   (8 vectors)
    $0000 +---------------+


  Naming:
	data block : struct { dw.size, data[size] } in malloced data area
	handle	   : struct { db.refcnt, dw.memptr } in handles area
	refcnt	   : reference count; 0=free
	memptr	   : pointer to data.data of a data block

	pointers in variables point to handle.memptr


  Provided VM Opcodes and MC Functions:

	mem_refcnt		( ptr -- u8 )
	mem_retain		( ptr -- ptr )
	mem_release		( ptr -- )
	mem_get_free	( -- u16 )
	mem_alloc	 	( size -- ptr )
	spacestr	 	( size filler -- str )
	istr			( [len,text] -- cstr )

	; assumes that heap_start and handles_end are on page boundary! (low byte = 0)
	garbage_collection ( -- )	VM opcode, uses stack
	mem_compact		( -- )		mcode entry


	assert_free_handle		( -- )
	create_free_handle		( -- )

	new_handle_for_ptr		( de->data -- de->handle )
	new_handle_for_ptr_fast	( de->data -- de->handle )
	new_handle_with_size	( de:size -- de->handle )		unused
	new_data_with_size		( de:size -- de->data )
	new_data_with_size_fast	( de:size -- de->data )
	run_mem_compact			( -- )
	oomem			( -- no return )
	memclr			( hl->data de:size a:filler -- hl->data de:0 a:filler )
	mem_compact		( -- )
.endif


; _________________________________________________
; initialized system variables

.area RAM
heap_start:		dw  dflt_heap	; -> start of heap
heap_end:		dw  dflt_heap	; -> end of heap
handles_start:	dw  ram_end	 	; -> start of memory handles
handles_end:  	dw  ram_end	 	; -> end of memory handles
							 	;    start of resident data, e.g. sio buffers
free_handle:  	dw  0		 	; -> last free handle or 0



.area CODE

; _________________________________________________
; get size of immediately free memory
; which can be allocated without running mem_compact()
; returned size is what actually can be allocated
; overhead already subtracted.
; note:	after system start required overhead may be 2 or 5 bytes,
;   	after a while running it is almost always 2 bytes only.
; note:	a return value of 0 does not guarantee that a request for 0 bytes
;   	will not trigger a GC or even will not abort with oomem!

mem_get_free::	; ( -- u16 )

	ld	hl,(handles_start)
	ld	de,(heap_end)

	; test free_handle:
	ld	a,(free_handle+1)	; test only the high byte
	and	a
	jr	z,1$

	dec	hl		; else subtract 3 for new handle
	dec	hl
	dec	hl

1$:	dec	hl
	scf
	sbc	hl,de	; hl = max_size
	ret	nc

; remaining space is negative => return 0
; note: we could also return hl as-is and declare the return value as a signed int.
; but this works only for heap size up to 32kB.

	ld	hl,0
	ret


; _________________________________________________
; get reference count for handle

mem_refcnt::	; ( ptr -- u8 )

	dec	hl		; hl -> handle.refcnt
	ld	l,(hl)
	ret


; _________________________________________________
; retain handle

mem_retain::	; ( ptr -- ptr )		mod: f, pres all other, may panic

	dec	hl
	inc	(hl)
	inc	hl
	ret	nz
	ld	a, REFCNT_OVERFLOW
	jp	panic	; jp => show ip of caller


; _________________________________________________
; release handle

mem_release::	; ( ptr -- )			mod: af de hl

	dec	hl		; hl->handle.refcnt
	dec	(hl)	; release handle
	ret	nz		; still in use

; store current free_handle as next free handle in handle hl's memptr
; store handle hl as new free_handle:

	inc	hl
	ld	de,(free_handle)
	ld	(free_handle),hl
	ld	(hl++),e
	ld	(hl),d
	ret


; _________________________________________________
; get an IVAL string
; the handle is retained and points to the const string in VM code
; releasing the handle will just free the handle
; note:
; getting an IVAL string multiple times creates a new handle each time.
; it is recommended to get it only once and then copy it as needed.

istr::			; ( [len,text] -- ptr )

	pop	de				; de = len
	ld	hl,0
	add	hl,sp			; hl -> text
	ex	hl,de			; de -> text, hl = len
	add	hl,de			; de -> text, hl -> behind text (may be odd)
	inc	hl
	res	0,l				; de -> text, hl -> behind text
	ld	sp,hl			; sp -> next opcode

	ld	a,(free_handle+1)	; test whether we have a free handle
	and	a
	jp	nz,new_handle_for_ptr_fast ; get handle and return to next command

	ld	(restore_sp+1),hl	; hl=sp
	ld	sp,iy
	call new_handle_for_ptr
	jp	restore_sp


; _________________________________________________
; allocate memory and clear it with fillbyte
; --> spacestr()

spacestr::		; ( size filler -- str )

	ld	 (restore_sp+1),sp
	ld	 sp,iy

	ld	 a,e
	push af	; filler
	push hl	; size

	call new_data_with_size		; ( hl:size -- hl->data )

	pop	 de			; size
	pop	 af			; filler
	and	 a
	call nz,memset_a				; ( hl->data de:size a:filler -- hl->data )

	call new_handle_for_ptr_fast	; ( de->data -- de->handle )  & resume VM
	jp	 restore_sp


; _________________________________________________
; allocate memory cleared with zero
; note: all memory in the heap is already zeroed

mem_alloc::		; ( size -- ptr )

	ld	(restore_sp+1),sp
	ld	sp,iy
	call new_data_with_size		; ( hl:size -- hl->data )
	call new_handle_for_ptr_fast
	jp	restore_sp




; -------------------------------------
;	mcode helpers:
; -------------------------------------



; _________________________________________________
; assert free handle available
; may panic with OOMEM
;
; assert that a free handle can be obtained from the linked list of free handles
; this proc does always a quick exit after the system was running for a while
; note: caller may do the the test itself and only call create_free_handle

assert_free_handle:			; ( -- )		mod: af de, pres: hl		uses stack		unused

	ld	a,(free_handle+1)	; test only the high byte
	and	a
	ret	nz
	;jr	create_free_handle

; _________________________________________________
; create free handle
; caller must assert that free_handle is actually null (see assert_free_handle)
; may panic with OOMEM

create_free_handle:			; ( -- )		mod: af de,  pres: hl		uses stack		used

	push hl
	ld	hl,(handles_start)	; hl = handles_start
	dec	hl
	dec	hl
2$:	scf
	ld	de,(heap_end)		; bc = heap_end
	sbc	hl,de				; hl = gap_size -3
	call c,1$				; less than 3 bytes free

	add	hl,de				; hl = handles_start-3 = new handles_start
	ld	(handles_start),hl	; update handles_start
	inc	hl
	ld	(free_handle),hl	; update free_handle

	; note:	new handle.refcnt = 0	because heap memory is always cleared to 0
	;	new handle.memptr = 0	""

	pop	hl
	ret

1$:	call mem_compact
	jr	nz,2$
	;jr	oomem

oomem::
	ld 	a, OUT_OF_MEMORY
	jp	panic


; _________________________________________________
; get free handle, update 'free_handle',
; store pointer hl in handle and return pointer to handle
; free handle must exist!

new_handle_for_ptr_fast::	; ( hl->data -- hl->handle )	mod: af de, 	no stack use

	ex	hl,de				; de -> data
	ld	hl,(free_handle)	; hl -> handle
	dec	hl
	inc	(hl)				; refcnt++
	inc	hl

	; load next free handle from memptr into de
	; and store address in de in memptr:
	; ex de,(hl):
	ld	a,(hl)
	ld	(hl),e
	ld	e,a
	inc	hl
	ld	a,(hl)
	ld	(hl),d
	ld	d,a    				; de -> next free handle
	dec	hl					; hl -> handle

	ld	(free_handle),de	; update next_free  (may be 0)
	ret


; _________________________________________________
; allocate hl bytes on heap
; may panic

new_handle_with_size:		; ( hl:size -- hl->handle )		mod: af de hl		unused

	ld	de, new_handle_for_ptr_fast
	push de
	;jr	new_data_with_size


; _________________________________________________
; allocate hl bytes on heap
; asserts that a free handle is available
; allocates data and returns address
; may panic

new_data_with_size::	; ( hl:size -- hl->data )	mod: af de hl			used

	ld	a,(free_handle+1)		; assert free handle: must be done before allocating data
	and	a						; else data might move if new_handle triggers a GC
	call z,create_free_handle	; pres. a hl
	;jr	new_data_with_size_fast


; _________________________________________________
; allocate hl bytes on heap
; a free handle must exist!
; may panic
;
; note: normally new_data_with_size should be used instead of new_data_with_size_fast.
;	the caller must assert that the garbage collection is not triggered
;	before the returned data address is stored in a handle, e.g. by calling assert_free_handle.
;	else we'd need to add +3 here if no free handle is available.

new_data_with_size_fast::	; ( hl:size -- hl->data )		mod f hl de, pres: a

	push hl

	;assert free memory:
	ld	hl,(handles_start)
	scf
	sbc	hl,de			; hl = handles_start - (size+1)
	jr	c, oomem

2$:	ld	de,(heap_end)
	scf
	sbc	hl,de			; hl = free_mem - (size+2)
	call c, 1$

	pop	de				; de = size
	ld	hl,(heap_end)
	ld	(hl++),e		; store size
	ld	(hl++),d		; hl -> data
	ex	hl,de			; de -> data, hl = size
	add	hl,de			; de -> data, hl -> behind data
	ld	(heap_end),hl
	ret

1$:	call mem_compact
	jr	nz, 2$
	jr	oomem
















; _________________________________________________
; get new handle
; ->	handle.memptr = address
; 	handle.refcnt = 1 (retained)
;
; note: if the data block is in the heap, then the caller must assert beforehand
; 	that enough memory is available for a new handle after allocating the data block
;	or that a free handle exists. Otherwise new_handle_for_ptr may trigger a GC
;	and the data pointer becomes invalid!
;
; in:  	de = memptr
; out: 	de = handle
; mod: 	af,de,hl
; pres:	bc

new_handle_for_ptr::	; ( de->data -- de->handle )

	ld	a,(free_handle+1)	; assert free handle
	and	a
	call	z,create_free_handle	; pres bc, de
	jr	new_handle_for_ptr_fast



; _________________________________________________
; run the garbage collection
; all registers preserved
; does not disable interrupts
; note:
; assumes that heap_start and handles_end are on page boundary! (low byte = 0)
; return: nz: some space freed
;		  z:  no space freed

garbage_collection::	; ( -- f:z ) 	mod: af,  pres: all others 		uses stack

	ld	 (restore_sp+1),sp
	ld	 sp,iy
	call mem_compact
	jp	restore_sp

; mcode entry:

mem_compact::
	push	hl
	push	de
	push	bc

.if 0
Da die Garbage Collection länger dauern kann, gibt es Probleme mit den Interrupts:
die müssen eigentlich gesperrt werden, weil auch die Interrupt-Handler auf Objekte im Heap zugreifen können.
Das kann umgangen werden, wenn alle Objekte für Interrupt-Handler beim Systemstart angelegt werden,
und somit ganz unten im Heap liegen, und nie wieder gelöscht oder neu zugewiesen werden.
Dann wird es in diesem Bereich keine Lücken geben und mem_compact() kann diesen Bereich ignorieren.
Tatsächlich scheint es sinnvoll zu sein, nach der System-Initialisierung heap_start entsprechend hochzusetzen. !!!
Andernfalls müssten wir erst einmal die erste Lücke suchen, damit wir wissen,
unterhalb welcher Adresse Speicher ignoriert werden kann.
.endif

; step 1:
; for all handles, if used and memory inside heap, then swap handle.memptr and mem.len

	ld	hl,(handles_start)	; hl -> first handle
	jr	4$

1$:	inc	hl
2$:	inc	hl

; test for end:
4$:	ld	a,l
	and	a
	jr	z,3$			; on page boundary => may be handles_end

; inspect refcnt:
5$:	ld	a,(hl++)		; a = refcnt, hl -> handle.memptr
	and	a
	jr	z,1$			; free handle => skip

; inspect memptr:
	ld	e,(hl++)
	ld	d,(hl)			; de = memptr, hl -> handle.memptr.hi
	bit	7,d
	jr	nz,2$			; memptr ≥ $8000 => in rom => skip

	ld	a,(heap_start+1)	; heap_start.hi
	scf
	sbc	a,d			; heap_start.h - 1 - memptr.hi
	jr	nc,2$			; nc =>  memptr.hi < heap_start.hi  =>  not in heap
	jr	nz,6$			; memptr.hi > heap_start.hi  =>  in heap

	; memptr.hi == heap_start.hi => kann drüber oder drunter sein => compare low bytes
	; note: heap_start liegt nicht auf page boundary, weil heap_start nach system init
	; über die "statischen" Speicherbereiche angehoben wurde
	ld	a,(heap_start)		; heap_start.lo
	;scf				; still set
	sbc	a,e			; heap_start.lo - 1 - memptr.lo
	jr	nc,2$			; nc =>  memptr.lo < heap_start.lo  =>  not in heap


; swap mem.len <-> handle.memptr
; de -> mem.len
; hl -> handle.memptr.hi

; get len: bc=(de) and store &memptr: (de)=hl
6$:	dec	hl			; hl -> memptr
	ex	hl,de			; de -> handle.memptr, hl -> mem.len
	ld	c,(hl)
	ld	(hl++),e
	ld	b,(hl)			; bc = len
	ld	(hl),d			; mem.len -> handle.memptr
	dec	hl
	ex	hl,de

; store len in handle.memptr and advance to next handle
	ld	(hl++),c
	ld	(hl++),b
	jr	4$


; hl on page boundary: check for handles_end:
3$:	ld	a,(handles_end+1)	; handles_end.hi
	cp	h
	jr	nz,5$			; resume


; step 2
; for all memory blocks in heap,
;   mem.len < $8000: always true (rom starts at $8000)
;   mem.len ≥ $4000 => a pointer: preserve block, move it down and swap memptr and len back
;   mem.len < $4000 => a length: they are unused. skip them, increasing the gap.
; there are two loops: the first is run while no gap was yet seen and memory block actually do not move.
; the second runs after the first gap has been seen.

	ld	hl,(heap_start)	; src

; loop: while no gap found:

11$:	ld	de,hl		; dest
	ld	c,(hl++)	;
	ld	b,(hl)		; bc = len, hl->mem.len.hi
	bit	6,b
	jr	z,13$		; < $4000 => a length => free block

; len ≥ $4000 => a used block => no need to move it down because there wasn't yet a gap:
; swap len and memptr back

; bc -> handle.memptr
; hl -> mem.len.hi
; de -> mem.len.lo

	ld	hl,bc		; hl -> handle.memptr, de -> mem.len

;	ld	bc,(hl)		; bc = len
;	ld	(hl),de		; handle.memptr -> mem.len.lo ✓
	ld	c,(hl)
	ld	(hl++),e
	ld	b,(hl)
	ld	(hl),d

	ex	hl,de		; hl -> mem.len

	ld	(hl++),c	; mem.len := len ✓
	ld	(hl++),b	; hl -> mem.data
	add	hl,bc		; hl -> next mem block

	ld	a,(heap_end+1)	; a = heap_end.hi
	cp	a,h
	jr	nz,11$		; not yet at heap_end
	ld	a,(heap_end)	; a = heap_end.lo
	cp	a,l
	jr	nz,11$		; not yet at heap_end

	jr	99$		; finished: no memory freed!


; loop: after a gap was found:

; hl -> next mem block (ldir source)
; de -> behind last mem block (ldir dest)

14$:	ld	c,(hl++)	;
	ld	b,(hl)		; bc = len, hl->mem.len.hi
	bit	6,b
	jr	z,13$		; < $4000 => a length => free block

; len ≥ $4000 => a used block => move it down
; swap len and memptr back

; bc -> handle.memptr
; hl -> mem.len.hi (source)
; de -> mem.len.lo (dest)

10$:	push	hl		; stack: ptr -> mem.len.hi (source)

	ld	hl,bc		; hl -> handle.memptr, de -> mem.len

;	ld	bc,(hl)		; bc = len
;	ld	(hl),de		; handle.memptr -> mem.len.lo ✓
	ld	c,(hl)
	ld	(hl++),e
	ld	b,(hl)
	ld	(hl),d

	pop	hl			; hl -> mem.len.hi (source), de -> mem.len.lo (dest), bc = len
	ld	a,c
	ld	(de++),a	; hl -> mem.len.hi (source), de -> mem.len.hi (dest), bc = len
	ld	(hl),b		; to be copied by ldir
	inc	bc			; hl -> mem.len.hi (source), de -> mem.len.hi (dest), bc = len+1

	ldir			; hl -> next mem block (source), de -> behind last mem block (dest)
	jr	14$


; len < $4000 => a length => free block => collapse it

; bc =  len
; hl -> mem.len.hi
; de -> mem.len.lo

13$:	inc	hl		; hl -> mem.data
	add	hl,bc		; hl -> next mem block
12$:	ld	a,(heap_end+1)
	cp	a,h
	jr	nz,14$
	ld	a,(heap_end)
	cp	a,l
	jr	nz,14$

; last block reached => update heap_end and exit
	ld	(heap_end),de


	abort	; TODO: clear freed memory with 0


99$:	pop	bc
	pop	de
	pop	hl
	; TODO: return z if no space reclaimed, nz otherwise
	ret


.endlocal

